/**
   \file FileTaggedMessageQueue.cpp

   \brief Fila de mensagens (com tag) utilizando um arquivo para leitura 
   e escrita.

   Copyright&copy; 1998-2017 Centro de Sensoriamento Remoto / Universidade
   Federal de Minas Gerais - Brazil.
   All Rights Reserved.

   \author Romulo Fernandes Machado
*/


#include "Config.h"
#define _DFF_CPP_LIBRARY_ BaseLib

#include <fstream>
#include <chrono>
#include <cstdio>
#include <cstring>
#include <thread>

#include "Exception.h"
#include "ExceptionUtils.h"

#include "FileTaggedMessageQueue.h"
#include "FilenameLock.h"
#include "Log.h"

#ifdef _DFF_ON_WINDOWS_
# define _WINDOWS_
#endif

#define BLOCK_SLEEP_INTERVAL_MS 100
#define LOCK_SUFFIX ".lock"


namespace DFF {


namespace {


/**
   \brief Permite o uso de variveis opcionais.

    necessrio declarar o prprio optional aqui uma vez que o cdigo do R
   depende desse arquivo e boost::optional no pode ser usado.
*/
template < typename T >
class Optional {
public:
  Optional() : m_initialized( false ) {}

  T value() const { 
    if ( !m_initialized )
      throwException( DFF::Exception, "Value is not initialized." );

    return m_value;
  }
    
  Optional& operator=( const T& rhs ) {
    m_value = rhs;
    m_initialized = true;

    return *this;
  }

  explicit operator bool() const noexcept { 
    return m_initialized;
  }

private:

  T m_value;
  bool m_initialized;
};


inline void writeHeader( std::fstream& stream, 
  const DFF::FileTaggedMessageQueue::Header& header ) {

  stream.seekp( 0 );
  stream.write( ( char* ) &header, sizeof( DFF::FileTaggedMessageQueue::Header ) );
  stream.flush();
}

template < typename T >
DFF::FileTaggedMessageQueue::Header getHeader( T& stream ) {
  DFF::FileTaggedMessageQueue::Header header;

  stream.seekg( 0 );
  stream.read( ( char* ) &header, sizeof( DFF::FileTaggedMessageQueue::Header ) );

  return header;
}

inline void writeMessageHeader( std::fstream& stream, 
  const DFF::FileTaggedMessageQueue::MessageHeader& header ) {

  stream.write( ( char* ) &header, sizeof( DFF::FileTaggedMessageQueue::MessageHeader ) );
  stream.flush();
}

inline void writeMessage( std::fstream& stream, 
  const DFF::FileTaggedMessageQueue::MessageHeader& header, const void* data, 
  size_t dataLength ) {

  writeMessageHeader( stream, header );
  stream.write( static_cast< const char* >( data ), dataLength );
  stream.flush();
}

template < typename T >
DFF::FileTaggedMessageQueue::MessageHeader getMessageHeader( T& stream ) {
  DFF::FileTaggedMessageQueue::MessageHeader header;

  stream.read( ( char* ) &header, sizeof( DFF::FileTaggedMessageQueue::MessageHeader ) );
  return header;
}

// Remove um arquivo de forma portvel.
// (Pending) Usar C++ 17 ou bibliotecas do boost para remoo do arquivo.
void removeFile( const std::string& filename ) {
  if ( std::remove( filename.c_str() ) != 0 )
    logDebug( "Failed to remove file \"" << filename << "\"" );
}


} // namespace ''


FileTaggedMessageQueue::FileTaggedMessageQueue( const std::string& filename, 
  size_t messageSize, size_t messageCount ) 
  : m_filename( filename ),
    m_messageSize( messageSize ),
    m_maximumMessageCount( messageCount ) { 

  // Obtm a trava exclusiva para criao do arquivo (como supe-se que 
  // o arquivo no existe, dispara se a trava existir).
  const std::string lockFilename( m_filename + LOCK_SUFFIX );
  const FilenameLock lock( lockFilename, FilenameLock::LB_THROW_ON_BUSY );

  // Verifica se a sesso j existe.
  if ( std::ifstream( m_filename, std::ios::in ).is_open() )
    throwException( DFF::Exception, "Failed to create tagged message "
      "queue \"" << m_filename << "\". Session already exist." );

  // Cria o arquivo em modo binrio.
  std::fstream handle( m_filename, std::ios::out | 
    std::ios::binary | std::ios::trunc );

  if ( handle.is_open() == false )
    throwException( DFF::Exception, "Failed to create tagged message "
      "queue \"" << m_filename << "\"." );

  // Escreve o cabealho com as informaes 
  const FileTaggedMessageQueue::Header header = { messageSize, messageCount };
  writeHeader( handle, header );

  // Escreve no arquivo bytes suficientes para acomodar todas as mensagens.
  const FileTaggedMessageQueue::MessageHeader msgHeader = {};

  char* tmpData = new char[ messageSize ];
  memset( tmpData, 0, messageSize );

  for ( size_t i = 0; i < messageCount; ++i ) {
    writeMessage( handle, msgHeader, tmpData, messageSize );
  }

  delete [] tmpData;
}

FileTaggedMessageQueue::FileTaggedMessageQueue( const std::string& filename ) 
  : m_filename( filename ) {

  // Obtm a trava exclusiva para leitura do arquivo.
  const std::string lockFilename( m_filename + LOCK_SUFFIX );
  const FilenameLock lock( lockFilename );

  std::ifstream handle( m_filename, std::ios::in | 
    std::ios::out | std::ios::binary );

  if ( handle.is_open() == false )
    throwException( DFF::Exception, "Failed to open tagged message "
      "queue \"" << m_filename << "\"." );

  // Recupera o cabealho para preencher o tamanho da mensagem armazenada.
  const FileTaggedMessageQueue::Header header( getHeader( handle ) );
  m_messageSize = header.blockSize;
  m_maximumMessageCount = header.blockCount;
}


void FileTaggedMessageQueue::write( const std::string& tag, const void* data, 
  size_t size, Cancellable* cancellable ) {

  while ( true ) {
    const std::string lockFilename( m_filename + ".lock" );
    FilenameLock lock( lockFilename );

    std::fstream handle( m_filename, std::ios::in | 
      std::ios::out | std::ios::binary );

    if ( handle.is_open() == false )
      throwException( DFF::Exception, "Failed to open tagged message "
        "queue \"" << m_filename << "\"." );

    const FileTaggedMessageQueue::Header header( getHeader( handle ) );
    const size_t messagesStartIndex = handle.tellg();

    // Procura o ndice da mensagem com o tag especificado ou o primeiro 
    // slot disponvel, caso a tag no aparea em nenhuma mensagem.
    Optional< size_t > messageSlot;
    for ( size_t i = 0; i < m_maximumMessageCount; ++i ) {
      handle.seekg( messagesStartIndex + i * ( m_messageSize + 
        sizeof ( FileTaggedMessageQueue::MessageHeader ) ) );

      const FileTaggedMessageQueue::MessageHeader msgHeader( 
        getMessageHeader( handle ) );

      // Uma vez que a tag  encontrada, a busca pode ser encerrada.
      if ( msgHeader.inUse && msgHeader.hasTag( tag ) ) {
        messageSlot = i;
        break;
      } 

      // Verifica se o slot est disponvel, caso a tag no seja encontrada.
      if ( !msgHeader.inUse && !messageSlot ) {
        messageSlot = i;
      }
    }

    // Foi encontrado um slot para gravao da mensagem ou uma
    // mensagem com a mesma tag existente.
    if ( messageSlot ) {
      handle.seekg( messagesStartIndex + messageSlot.value() * ( m_messageSize +
        sizeof ( FileTaggedMessageQueue::MessageHeader ) ) );

      writeMessage( handle, { true, tag }, data, size );
      break;
    }

    // Libera o lock e espera um timeout para uma nova tentativa.
    lock.release();
    std::this_thread::sleep_for( std::chrono::milliseconds( BLOCK_SLEEP_INTERVAL_MS ) );

    if ( cancellable != nullptr && cancellable->isCancelled() )
      throwException( DFF::Exception, "Execution aborted." );

  } // while ( true )
}


void FileTaggedMessageQueue::read( const std::string& tag, void* data, 
  size_t size, Cancellable* cancellable ) {
  while ( true ) {
    const std::string lockFilename( m_filename + ".lock" );
    FilenameLock lock( lockFilename );

    std::fstream handle( m_filename, std::ios::in | 
      std::ios::out | std::ios::binary );

    if ( handle.is_open() == false )
      throwException( DFF::Exception, "Failed to open tagged message "
        "queue \"" << m_filename << "\"." );

    const FileTaggedMessageQueue::Header header( getHeader( handle ) );
    const size_t messagesStartIndex = handle.tellg();
    for ( size_t i = 0; i < m_maximumMessageCount; ++i ) {
      handle.seekg( messagesStartIndex + i * ( m_messageSize + 
        sizeof ( FileTaggedMessageQueue::MessageHeader ) ) );

      const FileTaggedMessageQueue::MessageHeader msgHeader( 
        getMessageHeader( handle ) );

      if ( msgHeader.inUse && msgHeader.hasTag( tag ) ) {
        handle.read( ( char* ) data, size );
      
        // Aps ser lida, a mensagem deve ser marcada no arquivo.
        handle.seekg( messagesStartIndex + i * ( m_messageSize + 
          sizeof( FileTaggedMessageQueue::MessageHeader ) ) );

        writeMessageHeader( handle, { false, "\0" } );
        return;
      }
    }

    // Libera o lock e espera um timeout para uma nova tentativa.
    lock.release();
    std::this_thread::sleep_for( std::chrono::milliseconds( BLOCK_SLEEP_INTERVAL_MS ) );

    if ( cancellable != nullptr && cancellable->isCancelled() )
      throwException( DFF::Exception, "Execution aborted." );

  } // while ( true )
}

void FileTaggedMessageQueue::destroy( const std::string& filename ) {

  logDebug( "Destroying tagged file message queue \"" << ( filename + LOCK_SUFFIX ) << "\"" );
  removeFile( filename + LOCK_SUFFIX );

  logDebug( "Destroying tagged file message queue \"" << filename << "\"" );
  removeFile( filename );
}


} // namespace DFF
